from unittest import TestCase, main

import numpy as np

from tests.deterministic_data import (
    bound_random_walk,
    geometric_daily,
    geometric_hourly,
    mock_daily,
)


class TestDeterministicData(TestCase):
    def test_bound_random_walk(self):
        n_tests = 100

        with self.assertRaises(ValueError):
            bound_random_walk(-1, 100)

        with self.assertRaises(ValueError):
            bound_random_walk(100, -100)

        with self.assertRaises(TypeError):
            bound_random_walk(100, 200, 0.3)

        with self.assertRaises(ValueError):
            bound_random_walk(100, 200, -1)

        with self.assertRaises(ValueError):
            bound_random_walk(100, 200, -1, -0.1)

        with self.assertRaises(NotImplementedError):
            bound_random_walk(100, 100)

        for _ in range(n_tests):
            start = np.random.randint(1, 100)
            end = np.random.randint(1, 100)
            min_size = np.random.randint(100, 10000)

            if start == end:
                end += np.random.choice([-1, 1], 1)

            output = bound_random_walk(start, end, min_size=min_size, trend_strength=0.01)

            self.assertEqual(output[0], start)
            self.assertEqual(output[-1], end)

            if start > end:
                # should only hit the end price once
                self.assertEqual(np.sum(output <= end), 1)

            elif end > start:
                # should only hit the end price once
                try:
                    self.assertEqual(np.sum(output >= end), 1)
                except:
                    print(output, start, end)

    def test_mock_daily(self):
        n_tests = 100

        # check bad inputs
        with self.assertRaises(ValueError):
            mock_daily(-1, 100, 100)

        with self.assertRaises(ValueError):
            mock_daily(100, -100, 100)

        with self.assertRaises(ValueError):
            mock_daily(100, 100, -100)

        with self.assertRaises(TypeError):
            mock_daily(100, 100, 0.1)

        with self.assertRaises(NotImplementedError):
            mock_daily(100, 100, 100)

        # check that the start and end is always correct
        for _ in range(n_tests):
            start = np.random.randint(1, 100)
            end = np.random.randint(1, 100)
            n_days = np.random.randint(100, 1000)

            if start == end:
                end += np.random.choice([-1, 1], 1)

            output = mock_daily(start, end, n_days)

            # start and end are correct
            self.assertEqual(output.iloc[0], start)
            self.assertEqual(output.iloc[-1], end)
            # number of days is correct
            self.assertEqual(output.shape[0], n_days)
            # index is ordered correctly
            self.assertTrue(output.index.is_monotonic_increasing)
            # index is daily
            self.assertTrue(output.equals(output.resample("d").first()))

            # the start and end should be far appart
            while abs(end - start) / start < 0.5:
                start = np.random.randint(1, 100)
                end = np.random.randint(1, 100)
                n_days = np.random.randint(100, 1000)

            # check that the trend is strong
            output = mock_daily(start, end, n_days, trend_strength=10.0)

            # start and end are correct
            self.assertEqual(output.iloc[0], start)
            self.assertEqual(output.iloc[-1], end)
            # number of days is correct
            self.assertEqual(output.shape[0], n_days)
            # index is ordered correctly
            self.assertTrue(output.index.is_monotonic_increasing)
            # index is daily
            self.assertTrue(output.equals(output.resample("d").first()))
            # mean is between the start and end
            self.assertLessEqual(output.mean(), max(start, end))
            self.assertGreaterEqual(output.mean(), min(start, end))

    def test_geometric_daily(self):
        output = geometric_daily(1, 2.7048138294215285, 100)
        self.assertEqual(output.iloc[0], 1)
        self.assertAlmostEqual(output.iloc[-1], 1.01**100)
        self.assertEqual(output.shape[0], 100)
        # number of unique values should be 2, nan + R
        self.assertEqual(output.pct_change().round(5).unique().shape[0], 2)

        output = geometric_daily(1, 10, 100)
        self.assertEqual(output.iloc[0], 1)
        self.assertAlmostEqual(output.iloc[-1], 10)
        self.assertEqual(output.shape[0], 100)
        # number of unique values should be 2, nan + R
        self.assertEqual(output.pct_change().round(5).unique().shape[0], 2)

        output = geometric_daily(1, 0.5, 100)
        self.assertEqual(output.iloc[0], 1)
        self.assertAlmostEqual(output.iloc[-1], 0.5)
        self.assertEqual(output.shape[0], 100)
        # number of unique values should be 2, nan + R
        self.assertEqual(output.pct_change().round(5).unique().shape[0], 2)
        self.assertTrue(output.index.is_monotonic_increasing)
        # index is daily
        self.assertTrue(output.equals(output.resample("d").first()))

    def test_geometric_hourly(self):
        output = geometric_hourly(1, 2.7048138294215285, 100)
        self.assertEqual(output.iloc[0], 1)
        self.assertAlmostEqual(output.iloc[-1], 1.01**100)
        self.assertEqual(output.shape[0], 100)
        # number of unique values should be 2, nan + R
        self.assertEqual(output.pct_change().round(5).unique().shape[0], 2)

        output = geometric_hourly(1, 10, 100)
        self.assertEqual(output.iloc[0], 1)
        self.assertAlmostEqual(output.iloc[-1], 10)
        self.assertEqual(output.shape[0], 100)
        # number of unique values should be 2, nan + R
        self.assertEqual(output.pct_change().round(5).unique().shape[0], 2)

        output = geometric_hourly(1, 0.5, 100)
        self.assertEqual(output.iloc[0], 1)
        self.assertAlmostEqual(output.iloc[-1], 0.5)
        self.assertEqual(output.shape[0], 100)
        # number of unique values should be 2, nan + R
        self.assertEqual(output.pct_change().round(5).unique().shape[0], 2)
        self.assertTrue(output.index.is_monotonic_increasing)
        # index is daily
        self.assertFalse(output.equals(output.resample("d").first()))
        self.assertTrue(output.equals(output.resample("h").first()))


if __name__ == "__main__":
    main()



from unittest import TestCase, main

import numpy as np
import pandas as pd

from areport import Report
from tests.deterministic_data import geometric_daily, mock_daily


class TestReport(TestCase):
    def test_init(self):
        n_tests = 100

        # check that the frequency is correct when we subsample
        for _ in range(n_tests):
            rate = np.random.randint(2, 60)
            mock_pf = mock_daily(1, 2, 100).iloc[::rate]
            report = Report(mock_pf)
            self.assertTrue(report.dt_pf_values.equals(mock_pf))
            self.assertLessEqual(abs(report.freq - 365 / rate), 1)

        # try to create a report with values that doesn't start at one
        with self.assertRaises(ValueError):
            Report(mock_daily(2, 3, 100))

    def test_from_returns(self):
        n_tests = 100

        for _ in range(n_tests):
            mock_pf = mock_daily(1, np.random.uniform(low=0.1, high=100), 100)
            report = Report.from_returns(mock_pf.pct_change().fillna(0.0))
            self.assertLessEqual((report.dt_pf_values - mock_pf).abs().max(), 1e-9)
            self.assertEqual(report.freq, 365)

            # check when using timestamps instead of datetimes
            ts_mock_pf = mock_pf.copy()
            ts_mock_pf.index = [t.timestamp() for t in ts_mock_pf.index]
            report = Report(ts_mock_pf)
            self.assertTrue(report.dt_pf_values.equals(mock_pf))
            self.assertEqual(report.freq, 365)

    def test_from_balances(self):
        n_tests = 100

        for _ in range(n_tests):
            mock_pf = mock_daily(1, np.random.uniform(low=0.1, high=100), 100)
            report = Report.from_balances(mock_pf)
            self.assertLessEqual((report.dt_pf_values - mock_pf).abs().max(), 1e-9)
            self.assertEqual(report.freq, 365)

            # check when using timestamps instead of datetimes
            ts_mock_pf = mock_pf.copy()
            ts_mock_pf.index = [t.timestamp() for t in ts_mock_pf.index]
            report = Report(ts_mock_pf)
            self.assertTrue(report.dt_pf_values.equals(mock_pf))
            self.assertEqual(report.freq, 365)

        # check raising ValueError
        invalid_balances = [1, 2, 3]
        with self.assertRaises(ValueError):
            Report.from_balances(invalid_balances)

    def test_final_pnl(self):
        n_tests = 100

        # create a report using timestamps
        for _ in range(n_tests):
            end_val = np.random.randint(2, 10)
            mock_pf = mock_daily(1, end_val, 100)
            ts_mock_pf = mock_pf.copy()
            ts_mock_pf.index = [t.timestamp() for t in ts_mock_pf.index]

            report = Report(ts_mock_pf)
            self.assertTrue(report.dt_pf_values.equals(mock_pf))
            self.assertEqual(report.freq, 365)
            self.assertEqual(report.final_pnl, (end_val - 1))

        # create a report using datetimes
        for _ in range(n_tests):
            end_val = np.random.randint(2, 10)
            mock_pf = mock_daily(1, end_val, np.random.randint(10, 600))
            report = Report(mock_pf)
            self.assertTrue(report.dt_pf_values.equals(mock_pf))
            self.assertEqual(report.freq, 365)
            self.assertEqual(report.final_pnl, (end_val - 1))

    def test_annal_returns(self):
        mock_pf = mock_daily(1, 2, 365, start_time=1672531200)
        self.assertEqual(mock_pf.shape[0], 365)
        self.assertEqual(mock_pf.index[0], pd.to_datetime("2023-01-01"))
        self.assertEqual(mock_pf.index[-1], pd.to_datetime("2023-12-31"))
        report = Report(mock_pf)
        expected_series = pd.Series([1.0], index=[2023], name="return")
        self.assertTrue(report.annual_returns.equals(expected_series))

    def test_monthly_returns(self):
        mock_pf = mock_daily(1, 2, 31, start_time=1672531200)
        self.assertEqual(mock_pf.shape[0], 31)
        self.assertEqual(mock_pf.index[0], pd.to_datetime("2023-01-01"))
        self.assertEqual(mock_pf.index[-1], pd.to_datetime("2023-01-31"))
        report = Report(mock_pf)
        expected_series = pd.Series([1.0], index=["2023-1"], name="return")
        self.assertTrue(report.monthly_returns.equals(expected_series))

    def test_quarterly_returns(self):
        mock_pf = mock_daily(1, 2, 31 + 28 + 31, start_time=1672531200)
        self.assertEqual(mock_pf.shape[0], 31 + 28 + 31)
        self.assertEqual(mock_pf.index[0], pd.to_datetime("2023-01-01"))
        self.assertEqual(mock_pf.index[-1], pd.to_datetime("2023-03-31"))
        report = Report(mock_pf)
        expected_series = pd.Series([1.0], index=["2023-3"], name="return")
        self.assertTrue(report.quarterly_returns.equals(expected_series))

    def test_mtd_performance(self):
        mock_pf = geometric_daily(1, 2, 17 + 31, start_time=1671062400)
        mock_pf_daily_returns = mock_pf.pct_change()
        self.assertEqual(mock_pf.shape[0], 17 + 31)
        self.assertEqual(mock_pf.index[0], pd.to_datetime("2022-12-15"))
        self.assertEqual(mock_pf.index[-1], pd.to_datetime("2023-01-31"))
        report = Report(mock_pf)

        # check date
        self.assertTrue(report.mtd_performance.index[0] == pd.to_datetime("2022-12-31"))
        self.assertTrue(report.mtd_performance.index[-1] == pd.to_datetime("2023-01-31"))

        # test daily returns
        expected_series = mock_pf_daily_returns.loc["2023-01-01":"2023-01-31"].round(5)
        self.assertTrue(
            (report.mtd_performance + 1).pct_change().loc["2023-01-01":"2023-01-31"].round(5).equals(expected_series)
        )

        # test mtd returns
        num = 0
        for i in range(mock_pf.shape[0]):
            available_data = mock_pf.iloc[: i + 1]

            # expected return
            if mock_pf.index[i].day == 1:
                num = 1
            expected_return = (1 + mock_pf_daily_returns.iloc[-1]) ** (num) - 1
            num += 1

            # actual return
            report = Report(available_data)
            mtd_perf = report.mtd_performance

            self.assertAlmostEqual(mtd_perf.iloc[-1], expected_return)

    def test_ytd_performance(self):
        mock_pf = geometric_daily(1, 2, 17 + 31 + 28, start_time=1671062400)
        mock_pf_daily_returns = mock_pf.pct_change()
        self.assertEqual(mock_pf.shape[0], 17 + 31 + 28)
        self.assertEqual(mock_pf.index[0], pd.to_datetime("2022-12-15"))
        self.assertEqual(mock_pf.index[-1], pd.to_datetime("2023-02-28"))
        report = Report(mock_pf)

        # check date
        self.assertTrue(report.ytd_performance.index[0] == pd.to_datetime("2022-12-31"))
        self.assertTrue(report.ytd_performance.index[-1] == pd.to_datetime("2023-02-28"))

        # test daily returns
        expected_series = mock_pf_daily_returns.loc["2023-01-01":"2023-02-28"].round(5)
        self.assertTrue(
            (report.ytd_performance + 1).pct_change().loc["2023-01-01":"2023-02-28"].round(5).equals(expected_series)
        )

        # test ytd returns
        num = 0
        for i in range(mock_pf.shape[0]):
            available_data = mock_pf.iloc[: i + 1]

            # expected return
            if mock_pf.index[i].month == 1 and mock_pf.index[i].day == 1:
                num = 1
            expected_return = (1 + mock_pf_daily_returns.iloc[-1]) ** (num) - 1
            num += 1

            # actual return
            report = Report(available_data)
            ytd_perf = report.ytd_performance

            self.assertAlmostEqual(ytd_perf.iloc[-1], expected_return)


if __name__ == "__main__":
    main()




import numpy as np
import pandas as pd


def compute_max_dd(x, freq=1000000):
    if np.inf in x or -np.inf in x:
        raise ValueError("Data contains inf values")

    assert isinstance(x, np.ndarray)
    window = freq
    x = pd.DataFrame(x)
    x = np.cumprod(1 + x)
    # Calculate the max drawdown in the past window days for each day in the series.
    # Use min_periods=1 if you want to let the first 252 days data have an expanding window
    roll_max = x.rolling(window, min_periods=1).max()
    dd = x / roll_max - 1.0
    return dd.values


import json
import os

import numpy as np
import pandas as pd
from tabulate import tabulate


from .utils import compute_max_dd


class Report:
    def __init__(self, pf_values: pd.Series):
        """Generic class for calculating return stream metrics

        Args:
            pf_values (pd.Series): cumulative portfolio values starting at one
        """
        if pf_values.iloc[0] != 1.0:
            raise ValueError(f"Supplied pf_values should start at one got {pf_values.iloc[0]}!")

        self.pf_values = pf_values.copy()
        self.dt_pf_values = pf_values.copy()

        if type(self.pf_values.index) == pd.DatetimeIndex:
            self.pf_values.index = [t.timestamp() for t in self.pf_values.index]

        else:
            self.dt_pf_values.index = [pd.to_datetime(t, unit="s") for t in self.dt_pf_values.index]

        self.pf_returns = self.pf_values.pct_change()
        self.dt_pf_returns = self.dt_pf_values.pct_change()

        self.pf_returns.iloc[0] = 0
        self.dt_pf_returns.iloc[0] = 0

        self.initial_capital = self.pf_values.iloc[0]
        self.final_capital = self.pf_values.iloc[-1]

        self.freq = 31536000 // np.mean(np.diff(self.pf_values.index))

        self.log_dir = "logs"

    @classmethod
    def from_returns(cls, returns: pd.Series):
        """Creates this object using returns instead of portfolio values

        Args:
            returns (pd.Series): series contain the returns of the portfolio
        """
        assert type(returns) == pd.Series

        return cls((returns + 1).cumprod())

    @classmethod
    def from_balances(cls, balances: pd.Series, **kwargs):
        """Creates this object using balances instead of portfolio values

        Args:
            balances (pd.Series): series contain the balances of the portfolio

        Raises:
            ValueError: balances must be a pd.Series object
        """
        if not isinstance(balances, pd.Series):
            raise ValueError("Balances must be a pd.Series object")

        return cls(balances / balances.iloc[0], **kwargs)

    def get_metrics(self):
        """Return all the common metrics"""
        return {
            "final_pnl": self.final_pnl,
            "ytd_performance": self.ytd_performance.iloc[-1],
            "cagr": self.cagr,
            "sharpe": self.sharpe_ratio,
            "sortino": self.sortino_ratio,
            "arithmetic_sharpe": self.arithmetic_sharpe_ratio,
            "annualized_std": self.annualized_standard_deviation,
            "max_dd": self.max_drawdown,
            "calmar": self.calmar_ratio,
            "monthly_win_rate": self.win_rate,
            "monthly_neutral_rate": self.neutral_rate,
            "monthly_loss_rate": self.loss_rate,
            "adj_monthly_win_rate": self.adj_win_rate,
            "avg_monthly_win": self.avg_win,
            "avg_monthly_loss": self.avg_loss,
            "monthly_slugging_ratio": self.slugging_ratio,
            "avg_annual_turnover": self.avg_turnover,
        }

    def print_metrics(self):
        """Print all the common metrics to a table"""
        print(tabulate(pd.DataFrame(self.get_metrics(), index=[0]).T, tablefmt="fancy_grid"))

    def print_annual_returns(self):
        """Print all the annual returns to a table"""
        print(tabulate(self.annual_returns.to_frame(), tablefmt="fancy_grid", headers=["Year", "PnL"]))

    def print_monthly_returns(self, lookback: int = 12):
        """Print the last n monthly returns to a table"""
        print(
            tabulate(self.monthly_returns.iloc[-lookback:].to_frame(), tablefmt="fancy_grid", headers=["Month", "PnL"])
        )

    def print_quarterly_returns(self, lookback: int = 8):
        """Print the last n quarterly returns to a table"""
        print(
            tabulate(
                self.quarterly_returns.iloc[-lookback:].to_frame(), tablefmt="fancy_grid", headers=["Quarter", "PnL"]
            )
        )

    def metrics_to_csv(self, output_dir="", file_name: str = "metrics.csv"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        pd.DataFrame(self.get_metrics(), index=[0]).to_csv(
            os.path.join(self.log_dir, output_dir, file_name), index=None
        )

    def metrics_to_json(self, output_dir="", file_name: str = "metrics.json"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        with open(os.path.join(self.log_dir, output_dir, file_name), "w") as f:
            json.dump(self.get_metrics(), f)

    def annual_returns_to_csv(self, output_dir="", file_name: str = "annual_returns.csv"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        self.annual_returns.sort_index(ascending=False).to_frame().T.to_csv(
            os.path.join(self.log_dir, output_dir, file_name), index_label="year"
        )

    def monthly_returns_to_csv(self, lookback: int = 12, output_dir="", file_name: str = "monthly_returns.csv"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        self.monthly_returns.iloc[-lookback:].sort_index(ascending=False).to_frame().T.to_csv(
            os.path.join(self.log_dir, output_dir, file_name), index_label="month"
        )

    def quarterly_returns_to_csv(self, lookback: int = 8, output_dir="", file_name: str = "quarterly_returns.csv"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        self.quarterly_returns.iloc[-lookback:].sort_index(ascending=False).to_frame().T.to_csv(
            os.path.join(self.log_dir, output_dir, file_name), index_label="quarter"
        )

    @property
    def final_pnl(self):
        return (self.final_capital - self.initial_capital) / self.initial_capital

    @property
    def sharpe_ratio(self):
        return (self.cagr - 0.0) / (np.sqrt(self.freq) * np.std(self.pf_returns))

    @property
    def arithmetic_sharpe_ratio(self):
        return np.sqrt(self.freq) * np.mean(self.pf_returns) / np.std(self.pf_returns)

    @property
    def sortino_ratio(self):
        return (self.cagr - 0.0) / (np.sqrt(self.freq) * np.nanstd(self.pf_returns[self.pf_returns < 0]))

    @property
    def win_rate(self):
        return (self.monthly_returns > 0).sum() / self.monthly_returns.count()

    @property
    def neutral_rate(self):
        return (self.monthly_returns == 0).sum() / self.monthly_returns.count()

    @property
    def loss_rate(self):
        return (self.monthly_returns < 0).sum() / self.monthly_returns.count()

    @property
    def adj_win_rate(self):
        return (self.monthly_returns > 0).sum() / (self.monthly_returns.count() - (self.monthly_returns == 0).sum())

    @property
    def avg_win(self):
        return self.monthly_returns[self.monthly_returns > 0].mean()

    @property
    def avg_loss(self):
        return self.monthly_returns[self.monthly_returns < 0].mean()

    @property
    def slugging_ratio(self):
        return self.avg_win / abs(self.avg_loss)

    @property
    def avg_turnover(self):
        if hasattr(self, "annual_turnover"):
            avg_turnover = np.mean(self.annual_turnover)
        else:
            avg_turnover = None
        return avg_turnover

    @property
    def monthly_returns(self):
        monthly_returns = self.dt_pf_values.resample("ME").last()
        monthly_returns /= self.dt_pf_values.resample("ME").last().shift(1).fillna(1)
        monthly_returns -= 1
        monthly_returns.index = [f"{i.year}-{i.month}" for i in monthly_returns.index]
        monthly_returns.name = "return"
        return monthly_returns

    @property
    def quarterly_returns(self):
        quarterly_returns = self.dt_pf_values.resample("QE").last()
        quarterly_returns /= self.dt_pf_values.resample("QE").last().shift(1).fillna(1)
        quarterly_returns -= 1
        quarterly_returns.index = [f"{i.year}-{i.month}" for i in quarterly_returns.index]
        quarterly_returns.name = "return"
        return quarterly_returns

    @property
    def annual_returns(self):
        annual_returns = self.dt_pf_values.resample("YE").last()
        annual_returns /= self.dt_pf_values.resample("YE").last().shift(1).fillna(1)
        annual_returns -= 1
        annual_returns.index = [i.year for i in annual_returns.index]
        annual_returns.name = "return"
        return annual_returns

    @property
    def drawdown(self):
        return compute_max_dd(self.pf_returns.values)

    @property
    def max_drawdown(self):
        return compute_max_dd(self.pf_returns.values).min()

    @property
    def cagr(self):
        return (self.final_capital / self.initial_capital) ** (self.freq / self.pf_values.shape[0]) - 1

    @property
    def annualized_standard_deviation(self):
        return np.sqrt(self.freq) * np.std(self.pf_returns)

    @property
    def calmar_ratio(self):
        return self.cagr / abs(self.max_drawdown)

    @property
    def ytd_start(self):
        year_start = self.dt_pf_values.index[-1].replace(month=1, day=1, hour=0, minute=0, second=0) - pd.Timedelta(
            seconds=1
        )

        # find the available year start (whichever is earlier, last date of previous year or the start of the year)
        if self.dt_pf_values.index.min() < year_start:
            available_year_start = self.dt_pf_values.loc[:year_start].index[-1]
        else:
            available_year_start = year_start

        return available_year_start

    @property
    def ytd_performance(self):
        return self.dt_pf_values.loc[self.ytd_start :] / self.dt_pf_values.loc[self.ytd_start :].iloc[0] - 1

    @property
    def mtd_start(self):
        month_start = self.dt_pf_values.index[-1].replace(day=1, hour=0, minute=0, second=0) - pd.Timedelta(seconds=1)

        # find the available month start (whichever is earlier, last date of previous month or the start of the month)
        if self.dt_pf_values.index.min() < month_start:
            available_month_start = self.dt_pf_values.loc[:month_start].index[-1]
        else:
            available_month_start = month_start

        return available_month_start

    @property
    def mtd_performance(self):
        return self.dt_pf_values.loc[self.mtd_start :] / self.dt_pf_values.loc[self.mtd_start :].iloc[0] - 1

    def get_recent_performance(self, years=1, months=0, days=0, end_date=None):
        if end_date is None:
            end_date = self.dt_pf_values.index[-1]

        lookback_date = end_date - pd.DateOffset(years=years, months=months, days=days)
        return (
            self.dt_pf_values.loc[lookback_date:end_date] / self.dt_pf_values.loc[lookback_date:end_date].iloc[0] - 1
        )

    def get_recent_cagr(self, years=1, months=0, days=0, end_date=None):
        if end_date is None:
            end_date = self.dt_pf_values.index[-1]

        lookback_date = end_date - pd.DateOffset(years=years, months=months, days=days)
        return (
            self.dt_pf_values.loc[lookback_date:end_date].iloc[-1]
            / self.dt_pf_values.loc[lookback_date:end_date].iloc[0]
        ) ** (self.freq / self.dt_pf_values.loc[lookback_date:end_date].shape[0]) - 1


class ReportComparison(object):
    def __init__(
        self,
        report: Report,
        benchmark_reports: dict[str:Report],
        portfolio_name: str = "Portfolio",
    ):
        """Compare the portfolio to set of benchmarks

        Args:
            report (Report): The portfolio report
            benchmark_reports (dict[str: Report]): A dictionary of benchmark reports
            figsize (tuple): Figure size in matplotlib format, will be converted for plotly
            plot_backend (str): 'matplotlib' or 'plotly' or 'plotly.svg'
        """
        self.pf_report = report
        self.benchmarks = benchmark_reports
        self.portfolio_name = portfolio_name

        self.log_dir = "logs"

    @property
    def batting_averages(self):
        """This metric tell you how frequenty the portfolio has higher returns than the benchmark"""
        output = {}
        for benchmark, bm_report in self.benchmarks.items():
            difference = (self.pf_report.monthly_returns - bm_report.monthly_returns).dropna()
            output[benchmark] = (difference > 0).sum() / difference.count()

        return output

    @property
    def spearman_correlations(self):
        """Calculates Spearman correlations between portfolio returns and benchmark returns for each benchmark"""
        output = {}
        for benchmark, bm_report in self.benchmarks.items():
            corr = self.pf_report.pf_returns.corr(bm_report.pf_returns, method="spearman", min_periods=1)
            output[benchmark] = corr

        return output

    @property
    def pearson_correlations(self):
        """Calculates Pearson correlations between portfolio returns and benchmark returns for each benchmark"""
        output = {}
        for benchmark, bm_report in self.benchmarks.items():
            corr = self.pf_report.pf_returns.corr(bm_report.pf_returns, method="pearson", min_periods=1)
            output[benchmark] = corr

        return output

    @property
    def kendall_correlations(self):
        """Calculates Kendall correlations between portfolio returns and benchmark returns for each benchmark"""
        output = {}
        for benchmark, bm_report in self.benchmarks.items():
            corr = self.pf_report.pf_returns.corr(bm_report.pf_returns, method="kendall", min_periods=1)
            output[benchmark] = corr

        return output

    @property
    def betas(self):
        """Calculates beta values representing the sensitivity of portfolio returns to changes in benchmark returns
        for each benchmark"""
        output = {}
        for benchmark, bm_report in self.benchmarks.items():
            beta = self.pf_report.pf_returns.cov(bm_report.pf_returns, min_periods=1) / bm_report.pf_returns.var()
            output[benchmark] = beta

        return output

    @property
    def alphas(self):
        """Calculates Jensen's alpha between portfolio returns and benchmark returns for each benchmark"""
        output = {}
        for benchmark, bm_report in self.benchmarks.items():
            beta = self.pf_report.pf_returns.cov(bm_report.pf_returns, min_periods=1) / bm_report.pf_returns.var()
            jensen_alpha = self.pf_report.cagr - (beta * bm_report.cagr)
            output[benchmark] = jensen_alpha

        return output

    @staticmethod
    def performance_to_pct(performance):
        return performance * 100

    @staticmethod
    def drawdown_to_pct(drawdown, index):
        return pd.Series(drawdown.flatten() * 100, index=index)

    def get_benchmark_metrics(self):
        """Retrieves various benchmark metrics including batting averages, Pearson correlations, and beta values"""
        return pd.DataFrame(
            [self.batting_averages, self.pearson_correlations, self.alphas, self.betas],
            index=["batting_average", "pearson_correlation", "alpha", "beta"],
        ).T

    def get_benchmark_monthly_returns(self, months=5):
        """Provides monthly returns for the portfolio and benchmarks in recent years

        Args:
            months (int): number of months to show for monthly returns. Default is 5.
        """

        list_output = []
        benchmarks = [self.pf_report] + list(self.benchmarks.values())

        for report in benchmarks:
            monthly_ret = report.monthly_returns.iloc[-months:].to_dict()
            list_output.append(monthly_ret)

        return list_output

    def get_benchmark_annual_returns(self, years=5):
        """Provides annual returns for the portfolio and benchmarks in recent years

        Args:
            years (int): number of years to show for annual returns. Default is 5.
        """

        list_output = []
        benchmarks = [self.pf_report] + list(self.benchmarks.values())

        for report in benchmarks:
            annual_ret = report.annual_returns.iloc[-years:].to_dict()
            list_output.append(annual_ret)

        return list_output

    def get_benchmark_trailing_cagr(self):
        """Calculates the trailing compound annual growth rate (CAGR) for different time periods of portfolio and benchmarks"""
        list_output = []
        benchmarks = [self.pf_report] + list(self.benchmarks.values())

        for report in benchmarks:
            output_benchmark = {
                "10 years returns": report.get_recent_cagr(years=10),
                "5 years returns": report.get_recent_cagr(years=5),
                "3 years returns": report.get_recent_cagr(years=3),
                "1 year returns": report.get_recent_cagr(years=1),
                "6 months returns": report.get_recent_cagr(months=6),
                "3 months returns": report.get_recent_cagr(months=3),
            }
            list_output.append(output_benchmark)

        return list_output

    def get_metrics(self):
        metrics = pd.DataFrame(self.pf_report.get_metrics(), index=[self.portfolio_name])
        metrics = pd.concat(
            [metrics] + [pd.DataFrame(x.get_metrics(), index=[y]) for y, x in self.benchmarks.items()], axis=0
        )
        metrics = pd.concat([metrics, self.get_benchmark_metrics()], axis=1)
        return metrics

    def print_metrics(self):
        """Print the performance metrics for the portfolio and all the benchmarks"""
        print(
            tabulate(
                self.get_metrics().T,
                tablefmt="fancy_grid",
                headers=[self.portfolio_name] + [x for x in self.benchmarks.keys()],
            )
        )

    def metrics_to_csv(self, output_dir="", file_name: str = "metrics.csv"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        self.get_metrics().to_csv(os.path.join(self.log_dir, output_dir, file_name))

    def metrics_to_json(self, output_dir="", file_name: str = "metrics.json"):
        os.makedirs(os.path.join(self.log_dir, output_dir), exist_ok=True)

        with open(os.path.join(self.log_dir, output_dir, file_name), "w") as f:
            json.dump(self.get_metrics(), f)

    def print_benchmark_metrics(self):
        """Prints the benchmark metrics in a formatted table"""
        metrics = pd.DataFrame(self.get_benchmark_metrics(), index=[x for x in self.benchmarks.keys()])

        print(tabulate(metrics.T, tablefmt="fancy_grid", headers=[x for x in self.benchmarks.keys()]))

    def print_benchmark_monthly_returns(self):
        """Prints the benchmark monthly returns for the portfolio and all the benchmarks"""
        metrics = pd.DataFrame(
            self.get_benchmark_monthly_returns(), index=["Portfolio"] + [x for x in self.benchmarks.keys()]
        )

        print(tabulate(metrics.T, tablefmt="fancy_grid", headers=["Portfolio"] + [x for x in self.benchmarks.keys()]))

    def print_benchmark_annual_returns(self):
        """Prints the benchmark annual returns for the portfolio and all the benchmarks"""
        metrics = pd.DataFrame(
            self.get_benchmark_annual_returns(), index=[self.portfolio_name] + [x for x in self.benchmarks.keys()]
        )

        print(
            tabulate(
                metrics.T, tablefmt="fancy_grid", headers=[self.portfolio_name] + [x for x in self.benchmarks.keys()]
            )
        )

    def print_benchmark_trailing_cagr(self):
        """Print the trailing cagr for the portfolio and all the benchmarks"""
        metrics = pd.DataFrame(
            self.get_benchmark_trailing_cagr(), index=[self.portfolio_name] + [x for x in self.benchmarks.keys()]
        )

        print(
            tabulate(
                metrics.T, tablefmt="fancy_grid", headers=[self.portfolio_name] + [x for x in self.benchmarks.keys()]
            )
        )